#!/usr/bin/env python2
#
# fstab.py
#
# (c) Copyright 2010 Michael Towers (larch42 at googlemail dot com)
#
# This file is part of the larch project.
#
#    larch is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    larch is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with larch; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#----------------------------------------------------------------------------
# 2010.11.05

import os, backend

def generate(partlist):
    """Build a suitable /etc/fstab for the newly installed system.
    partlist is a list of (partitions, mount-point) pairs, where the
    partition is supplied either as label or as device path.
    The file-system type - and UUID if necessary - can be got using blkid.
    """
    fstab = ('# fstab generated by larchin\n'
            '#<file system>   <dir>       <type>      <options>    <dump> <pass>\n\n')

    fstab += ('none            /dev/pts    devpts      defaults        0     0\n'
            'none            /dev/shm    tmpfs       defaults        0     0\n\n')

    # Go through the list adding lines for each entry, but first sort the
    # mount-points alphabetically
    plist = [(p[1], p[0]) for p in partlist]
    plist.sort()
    swaps = []
    parts = []
    for p, m in partlist:
        if p.startswith('/dev/'):
            # use blkid to get uuid
            uuid = run_capture('blkid', '-s', 'UUID', '-o', 'value', p)[1].strip()
            line = 'UUID=' + uuid

        else:
            line = 'LABEL=' + p
            p = run_capture('blkid', '-L', p)[1].strip()

        fstype = run_capture('blkid', '-s', 'TYPE', '-o', 'value', p)[1].strip() or '???'
        if fstype == 'swap':
            swaps.append(line + 'swap       swap   defaults        0     0\n')
        else:
            pas = '1' if (m == '/') else '2'
            parts.append('%-15s %-12s %-8s defaults,noatime  0  %s\n'
                    % (line, m, fstype, pas))

    for line in parts:
        fstab += line
    fstab += '\n# Swaps\n'
    for line in swaps:
        fstab += line
    fstab += '\n# Other partitions\n'

    # then go through the other non-removable, unmounted partitions only
    # include them ('noauto') if they have an acceptable fstype
    # and add to a list of mount-points in /mnt
    mnt_list = []
    for line in run_capture('blkid', '-c', '/dev/null', '-o', 'list')[1].splitlines():
        if line.startswith('/dev/sd'):
            i = line.find('(not mounted)')
            if i > 0:
                # Try to get label for unmounted devices only
                ls = line[:i].split(None, 2)
                opts = getopts(ls[1])   # options based on fs-type
                if not opts:
                    continue
                dev = os.path.basename(ls[0])
                if len(ls) < 3:     # no label
                    # Use device path
                    ln = ls[0]
                    # and the base-name as mount-point in /mnt
                    lm = dev

                else:
                    # Use label
                    ln = 'LABEL=' + ls[2]
                    # also as mount-point in /mnt
                    lm = ln

                # Only add non-removable devices
                if readfile('/sys/block/%s/removable' % dev[:3]).strip() == '0':
                    mnt_list.append('/mnt/' + dev)
                    fstab += ('%-15s /mnt/%-7s %s,noauto,users  0  0\n'
                            % (ln, dev, opts))

    return (fstab, mnt_list)


def getopts(fstype):
    opt = ' defaults'

    if fstype == 'ntfs':
        fstype = 'ntfs-3g'
        opt += ',umask=111,dmask=000'

    elif fstype == 'vfat':
        opt += ',umask=111,dmask=000'

    elif fstype == 'swap':
        return None

    else:
        opt += ',noatime'

    return fstype + opt



if __name__ == '__main__':
    class options:
        slave = False
        force = False
        quiet = False
    backend.init(options)
    if backend.query_yn(_("Try to generate fstab?")):
        f, m = generate([('Arch', '/'), ('home', '/home'), ('swap', None)])
        print m
        print '--------------------------------'
        print f
    sys_quit(0)
